from __future__ import annotations

from confluent_kafka import KafkaError  # type: ignore[import-untyped]


class PartitionMetadata(object):
    """
    Provides information about a Kafka partition.

    This class is typically not user instantiated.
    """

    # The dash in "-topic" and "-error" is needed to circumvent a
    # Sphinx issue where it tries to reference the same instance variable
    # on other classes which raises a warning/error.

    def __init__(
        self,
        id: int,  # noqa: A002
        leader: int = -1,
        replicas: list[int] | None = None,
        isrs: list[int] | None = None,
        error: KafkaError | None = None,
    ):
        self.id = id
        """Partition Id"""
        self.leader = leader
        """Current leader broker for this partition, or -1."""
        self.replicas = replicas or []
        """List of replica broker ids for this partition."""
        self.isrs = isrs or []
        """List of in-sync-replica broker ids for this partition."""
        self.error = error
        """Partition error, or None. Value is a KafkaError object."""

        # If default args were provided then initialize the Partition with a single replica
        if leader == -1 and replicas is None and isrs is None:
            self.leader = 1
            self.replicas = [1]
            self.isrs = [1]

    def __str__(self):
        return self.id

    def __len__(self):
        return len(self.replicas)



"""
mock_topics = {
    'sample_topic':{
        1: [],
        2: [],
        3: [],
        4: []
    }
}

offset_store = {
    'sample_topic_1': {
        'first_offset': 0,
        'next_offset': 0,
    }
}
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any

from confluent_kafka import KafkaException  # type: ignore[import-untyped]

from .message import Message

mock_topics: dict[str, dict[int, list[Message]]] = {}

offset_store: dict[str, dict[str, int]] = {}

__all__ = ["KafkaStore"]


class SingletonMeta(type):
    _instances: dict[SingletonMeta, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances or "clean" in kwargs.keys():
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class KafkaStore(metaclass=SingletonMeta):
    """
    In memory kafka store
    """

    FIRST_OFFSET = "first_offset"
    NEXT_OFFSET = "next_offset"

    def __init__(self, clean: bool = False) -> None:
        if clean:
            mock_topics.clear()
            offset_store.clear()

    @staticmethod
    def is_topic_exist(topic: str) -> bool:
        return topic in mock_topics.keys()

    @classmethod
    def is_partition_exist_on_topic(cls, topic: str, partition_num: int) -> bool:
        if not cls.is_topic_exist(topic=topic):
            raise KafkaException("Topic Does not exist")

        return mock_topics[topic].get(partition_num, None) is not None

    @staticmethod
    def get_number_of_partition(topic: str) -> int:
        return len(mock_topics[topic].keys())

    @staticmethod
    def create_topic(topic: str) -> None:
        if mock_topics.get(topic, None) is not None:
            raise KafkaException(f"{topic} exist is fake kafka")

        mock_topics[topic] = {}

    def create_partition(self, topic: str, partitions: int) -> None:
        if not self.is_topic_exist(topic=topic):
            self.create_topic(topic=topic)

        len_of_current_partition = len(mock_topics[topic].keys())
        if partitions >= len_of_current_partition:
            for i in range(len_of_current_partition, partitions):
                mock_topics[topic][i] = []
                offset_store[self.get_offset_store_key(topic, i)] = {
                    self.FIRST_OFFSET: 0,
                    self.NEXT_OFFSET: 0,
                }

        else:
            raise KafkaException("can not decrease partition of topic")

    def remove_topic(self, topic: str) -> None:
        if not self.is_topic_exist(topic=topic):
            return

        mock_topics.pop(topic)

        offset_store_keys = deepcopy(list(offset_store.keys()))
        for offset_key in offset_store_keys:
            if topic in offset_key:
                offset_store.pop(offset_key)

    def set_first_offset(self, topic: str, partition: int, value: int) -> None:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        first_offset = self.get_partition_first_offset(topic=topic, partition=partition)
        next_offset = self.get_partition_next_offset(topic=topic, partition=partition)

        if first_offset < value <= next_offset:
            offset_store[offset_store_key][self.FIRST_OFFSET] = value

    def _add_next_offset(self, topic: str, partition: int) -> None:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        offset_store[offset_store_key][self.NEXT_OFFSET] += 1

    def get_offset_store_key(self, topic: str, partition: int) -> str:
        return f"{topic}*{partition}"

    def produce(self, message: Message, topic: str, partition: int) -> None:
        if not topic:
            return

        if partition is None:
            raise KafkaException(
                "you must assign partition when you want to produce message"
            )

        if not self.is_topic_exist(topic=topic):
            if partition == 0:
                partition = 1
            self.create_partition(topic=topic, partitions=partition)

        if mock_topics[topic].get(partition, None) is None:
            raise KafkaException(
                f"can not produce on partition {partition} of {topic}, partition does not exist"
            )

        # add message to topic
        mock_topics[topic][partition].append(message)

        self._add_next_offset(topic=topic, partition=partition)

    def get_message(self, topic: str, partition: int, offset: int) -> Message:
        return self.get_messages_in_partition(topic=topic, partition=partition)[offset]

    def get_partition_first_offset(self, topic: str, partition: int) -> int:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        return offset_store[offset_store_key][self.FIRST_OFFSET]

    def get_partition_next_offset(self, topic: str, partition: int) -> int:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        return offset_store[offset_store_key][self.NEXT_OFFSET]

    @staticmethod
    def topic_list() -> list[str]:
        return list(mock_topics.keys())

    @staticmethod
    def partition_list(topic: str) -> list[int]:
        return list(mock_topics[topic].keys())

    @staticmethod
    def get_messages_in_partition(topic: str, partition: int) -> list[Message]:
        return mock_topics[topic][partition]

    def number_of_message_in_topic(self, topic: str) -> int:
        count_of_messages = 0
        for partition in self.partition_list(topic=topic):
            count_of_messages += len(
                self.get_messages_in_partition(topic=topic, partition=partition)
            )

        return count_of_messages

    def clear_topic_messages(self, topic: str) -> None:
        for partition in self.partition_list(topic=topic):
            self.clear_partition_messages(topic=topic, partition=partition)

    @staticmethod
    def clear_partition_messages(topic: str, partition: int) -> None:
        mock_topics[topic][partition] = []

    def reset_offset(self, topic: str, strategy: str = "latest") -> None:
        for partition in self.partition_list(topic=topic):
            key = self.get_offset_store_key(topic, partition)

            if strategy == "latest":
                offset_store[key][self.FIRST_OFFSET] = offset_store[key][
                    self.NEXT_OFFSET
                ]

            elif strategy == "earliest":
                offset_store[key][self.FIRST_OFFSET] = 0

    @staticmethod
    def fresh() -> None:
        mock_topics.clear()
        offset_store.clear()



from __future__ import annotations

import time
from typing import Any, Optional

from confluent_kafka import (  # type: ignore[import-untyped]
    TIMESTAMP_CREATE_TIME,
    TIMESTAMP_LOG_APPEND_TIME,
    TIMESTAMP_NOT_AVAILABLE,
    KafkaError,
)


class Message:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._headers: Optional[list[tuple[str, Optional[bytes]]]] = kwargs.get("headers", None)
        self._key: Optional[bytes] = kwargs.get("key", None)
        self._value: Optional[bytes] = kwargs.get("value", None)
        self._topic: Optional[str] = kwargs.get("topic", None)
        self._offset: Optional[int] = kwargs.get("offset", None)
        self._error: Optional[KafkaError] = kwargs.get("error", None)
        self._latency: Optional[float] = kwargs.get("latency", None)
        self._leader_epoch: Optional[int] = kwargs.get("leader_epoch", None)
        self._partition: Optional[int] = kwargs.get("partition", None)
        self._timestamp: int = kwargs.get("timestamp") or int(time.time() * 1000)
        self._timestamp_type = kwargs.get("timestamp_type", TIMESTAMP_CREATE_TIME)
        self._broker_receive_time = kwargs.get("broker_receive_time") or int(
            time.time() * 1000
        )

    def offset(self, *args, **kwargs):
        return self._offset

    def latency(self, *args, **kwargs):
        return self._latency

    def leader_epoch(self, *args, **kwargs):
        return self._leader_epoch

    def headers(self) -> Optional[list[tuple[str, Optional[bytes]]]]:
        return self._headers

    def key(self, *args, **kwargs) -> Optional[bytes]:
        return self._key

    def value(self, *args, **kwargs) -> Optional[bytes]:
        return self._value

    def timestamp(self, *args, **kwargs) -> tuple[int, int]:
        ts_info: tuple[int, int]
        if self._timestamp_type == TIMESTAMP_NOT_AVAILABLE:
            ts_info = (TIMESTAMP_NOT_AVAILABLE, 0)
        elif self._timestamp_type == TIMESTAMP_LOG_APPEND_TIME:
            ts_info = (TIMESTAMP_LOG_APPEND_TIME, self._broker_receive_time)
        else:
            ts_info = (self._timestamp_type, self._timestamp)
        return ts_info

    def topic(self, *args, **kwargs):
        return self._topic

    def partition(self, *args, **kwargs):
        return self._partition

    def error(self):
        return self._error

    def set_headers(self, *args, **kwargs):  # real signature unknown
        pass

    def set_key(self, *args, **kwargs):  # real signature unknown
        pass

    def set_value(self, *args, **kwargs):  # real signature unknown
        pass



from __future__ import annotations

from .aiokafka_admin_client import FakeAIOKafkaAdmin
from .aiokafka_consumer import FakeAIOKafkaConsumer
from .aiokafka_producer import FakeAIOKafkaProducer

__all__ = ["FakeAIOKafkaProducer", "FakeAIOKafkaAdmin", "FakeAIOKafkaConsumer"]



from __future__ import annotations

from typing import Optional, Union

from typing_extensions import LiteralString

from mockafka.cluster_metadata import ClusterMetadata
from mockafka.kafka_store import KafkaStore
from mockafka.message import Message

__all__ = ["FakeProducer"]


def _serialise(obj: object, name: LiteralString) -> Optional[bytes]:
    if obj is None or isinstance(obj, bytes):
        return obj
    if isinstance(obj, str):
        return obj.encode()

    # Match apparent behaviour of `confluent_kafka.Producer`
    raise ValueError(f"{name} must be str, bytes or None, not {type(obj)!r}")


class FakeProducer(object):
    def __init__(self, config: dict | None = None):
        self.kafka = KafkaStore()

    def produce(
        self,
        topic,
        value: Union[str, bytes, None] = None,
        key: Union[str, bytes, None] = None,
        partition=None,
        callback=None,
        on_delivery=None,
        timestamp=None,
        headers: Union[
            # While Kafka itself supports only list[tuple[...]], confluent_kafka
            # allows passing in a dict here.
            dict[str, Optional[bytes]],
            list[tuple[str, Optional[bytes]]],
            None,
        ] = None,
        **kwargs,
    ) -> None:
        if isinstance(headers, dict):
            headers = list(headers.items())
        # create a message and call produce kafka
        message = Message(
            topic=topic,
            value=_serialise(value, "value"),
            key=_serialise(key, "key"),
            partition=partition,
            callback=callback,
            on_delivery=on_delivery,
            timestamp=timestamp,
            headers=headers,
            **kwargs,
        )
        self.kafka.produce(message=message, topic=topic, partition=partition)

    def list_topics(self, topic=None, *args, **kwargs):
        return ClusterMetadata(topic)

    def abort_transaction(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def begin_transaction(self):
        # This method Does not support in mockafka
        pass

    def commit_transaction(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def flush(self, timeout=None):
        # This method Does not support in mockafka
        return 0

    def init_transactions(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def poll(self, timeout=None):
        # This method Does not support in mockafka
        return 0

    def purge(self, in_queue=True, *args, **kwargs):
        # This method Does not support in mockafka
        pass

    def send_offsets_to_transaction(self, positions, group_metadata, timeout=None):
        # This method Does not support in mockafka
        pass



from __future__ import annotations

import inspect
from functools import wraps

from mockafka.consumer import FakeConsumer


def consume(topics: list[str], auto_commit: bool = True):
    """
    A decorator for simulating message consumption using a FakeConsumer.

    Parameters:
    - topics (list[str]): A list of topics to subscribe to.
    - auto_commit (bool): Whether to automatically commit offsets after consuming messages.

    Example Usage:
    ```python
    @consume(topics=['test_topic'], auto_commit=False)
    def test_function(message):
        # Your test logic for processing the consumed message here
    ```
    """

    def decorator(func):
        # Get the function signature
        sig = inspect.signature(func)

        # Create a new signature with default values for message parameters
        new_params = []
        for param_name, param in sig.parameters.items():
            if param_name == 'message' and param.default == inspect.Parameter.empty:
                # Add default value to prevent pytest from treating it as a fixture
                new_param = param.replace(default=None)
                new_params.append(new_param)
            else:
                new_params.append(param)

        new_sig = sig.replace(parameters=new_params)

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Create a FakeConsumer instance and subscribe to specified topics
            fake_consumer = FakeConsumer()
            fake_consumer.subscribe(topics=topics)

            # Simulate message consumption
            while True:
                message = fake_consumer.poll()

                # Break if no more messages
                if message is None:
                    break

                # Call the original function with the consumed message
                func(message=message, *args, **kwargs)  # noqa: B026  # TODO: fix unpacking

            # Call the original function without a message parameter
            result = func(message=None, *args, **kwargs)  # noqa: B026  # TODO: fix unpacking
            return result

        # Apply the new signature to the wrapper
        wrapper.__signature__ = new_sig
        return wrapper

    return decorator



from __future__ import annotations

from functools import wraps
from typing import Awaitable, Callable, Optional, TypeVar

from typing_extensions import ParamSpec

from mockafka.aiokafka import FakeAIOKafkaProducer

P = ParamSpec('P')
R = TypeVar('R')


def aproduce(
    *,
    topic: str,
    value: Optional[bytes] = None,
    key: Optional[bytes] = None,
    partition: int = 0,
    headers: Optional[list[tuple[str, Optional[bytes]]]] = None,
) -> Callable[
    [Callable[P, Awaitable[R]]],
    Callable[P, Awaitable[R]],
]:
    """
    aproduce is a decorator for simulating message production using a FakeAIOKafkaProducer.

    It extracts the following parameters from the decorator args:

    - topic (str): Topic to produce to
    - value (str): Message value
    - key (str): Message key
    - headers (dict): Message headers
    - partition (int): Topic partition to produce to

    The decorator creates a FakeAIOKafkaProducer instance and calls send()
    to produce a message with the provided parameters.

    It then calls the decorated function.

    This allows you to test message production logic using the fake producer.

    Example usage:

    @aproduce(topic='test', value='foo')
    async def produce_message():
      # test logic here

    """

    def decorator(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Create a FakeProducer instance and produce the message
            fake_producer = FakeAIOKafkaProducer()
            await fake_producer.send(
                topic=topic, partition=partition, value=value, key=key, headers=headers
            )

            # Call the original function
            result = await func(*args, **kwargs)
            return result

        return wrapper

    return decorator



from __future__ import annotations

from .admin_client import FakeAdminClientImpl
from .consumer import FakeConsumer
from .decorators import (
    aconsume,
    aproduce,
    asetup_kafka,
    bulk_produce,
    consume,
    produce,
    setup_kafka,
)
from .message import Message
from .producer import FakeProducer

__all__ = [
    "FakeProducer",
    "FakeConsumer",
    "FakeAdminClientImpl",
    "Message",
    "produce",
    "bulk_produce",
    "setup_kafka",
    "consume",
    "asetup_kafka",
    "aconsume",
    "aproduce",
]



from __future__ import annotations

from functools import wraps
from typing import Awaitable, Callable, TypeVar

from aiokafka.admin import NewTopic  # type: ignore[import-untyped]
from typing_extensions import ParamSpec

from mockafka.aiokafka import FakeAIOKafkaAdmin
from mockafka.decorators.typing import TopicConfig

P = ParamSpec('P')
R = TypeVar('R')


def asetup_kafka(
    topics: list[TopicConfig],
    clean: bool = False,
) -> Callable[
    [Callable[P, Awaitable[R]]],
    Callable[P, Awaitable[R]],
]:
    """
    asetup_kafka is a decorator for setting up mock Kafka topics using a FakeAIOKafkaAdminClient.

    It takes the following parameters:

    - topics (list[dict]): List of topic configs, each containing:
        - topic (str): Topic name
        - partition (int): Partition count
    - clean (bool): Whether to clean existing topics first. Default False.

    The decorator creates a FakeAIOKafkaAdminClient instance and uses it to create
    the specified topics.

    This allows you to setup mock Kafka topics and partitions at test setup.

    Example usage:

    @asetup_kafka(topics=[{'topic': 'test', 'partitions': 1}])
    async def test_function():
      # test logic

    """

    def decorator(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Create a FakeAdminClient instance with the specified clean option
            fake_admin = FakeAIOKafkaAdmin(clean=clean)

            # Create specified topics using the FakeAdminClient
            for item in topics:
                topic = item.get("topic", None)
                partition = item.get("partition", None)
                await fake_admin.create_topics(
                    new_topics=[
                        NewTopic(
                            name=topic, num_partitions=partition, replication_factor=1
                        )
                    ]
                )

            # Call the original function
            result = await func(*args, **kwargs)
            return result

        return wrapper

    return decorator



from __future__ import annotations

import asyncio
from typing import Any, Optional

from aiokafka.util import create_future  # type: ignore[import-untyped]
from typing_extensions import LiteralString, Self

from mockafka.kafka_store import KafkaStore
from mockafka.message import Message


def _check_type(obj: object, name: LiteralString) -> None:
    if not (
        obj is None or isinstance(obj, bytes)
    ):
        # Match apparent behaviour of `aiokafka.AIOKafkaProducer`
        raise ValueError(f"{name} must be bytes or None, not {type(obj)}")


class FakeAIOKafkaProducer:
    """
    FakeAIOKafkaProducer is a mock implementation of aiokafka's AIOKafkaProducer.

    It allows mocking a kafka producer for testing purposes.

    Parameters:
    - args, kwargs: Passed to superclass init, not used here.

    Attributes:
    - kafka: KafkaStore instance for underlying storage.

    Methods:
    - _produce(): Create a Message and produce to KafkaStore.
      Takes topic, value, and optional partition.
    - start(): No-op.
    - stop(): No-op.
    - send(): Call _produce() with kwargs.
    - send_and_wait(): Call send().
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.kafka = KafkaStore()

    async def _produce(
        self,
        topic: str,
        value: Optional[bytes],
        key: Optional[bytes],
        partition: int,
        timestamp_ms: Optional[int],
        headers: Optional[list[tuple[str, Optional[bytes]]]] = None,
    ) -> None:
        _check_type(value, "value")
        _check_type(key, "key")

        # create a message and call produce kafka
        message = Message(
            topic=topic,
            value=value,
            key=key,
            partition=partition,
            timestamp_ms=timestamp_ms,
            headers=headers,
        )
        self.kafka.produce(message=message, topic=topic, partition=partition)

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass

    async def send(
        self,
        topic: str,
        value: Optional[bytes] = None,
        key: Optional[bytes] = None,
        partition: int = 0,
        timestamp_ms: Optional[int] = None,
        headers: Optional[list[tuple[str, Optional[bytes]]]] = None,
    ) -> asyncio.Future[None]:
        await self._produce(
            topic=topic,
            value=value,
            key=key,
            partition=partition,
            headers=headers,
            timestamp_ms=timestamp_ms,
        )
        future: asyncio.Future[None] = create_future()
        future.set_result(None)
        return future

    async def send_and_wait(
        self,
        topic: str,
        value: Optional[bytes] = None,
        key: Optional[bytes] = None,
        partition: int = 0,
        timestamp_ms: Optional[int] = None,
        headers: Optional[list[tuple[str, Optional[bytes]]]] = None,
    ) -> None:
        future = await self.send(
            topic=topic,
            value=value,
            key=key,
            partition=partition,
            headers=headers,
            timestamp_ms=timestamp_ms,
        )
        return await future

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exception_type: object,
        exception: object,
        traceback: object,
    ) -> None:
        await self.stop()
